/*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*
                              <vendorlogger.c>
DESCRIPTION

EXTERNALIZED FUNCTIONS

This software is contributed or developed by KYOCERA Corporation.
(C) 2017 KYOCERA Corporation

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 and
only version 2 as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.

*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*/
#define pr_fmt(fmt) "vendorlogger: " fmt
/*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*
 * INCLUDE FILES
*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*/
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/kcjlog.h>
#include <linux/errno.h>
#include <linux/rtc.h>

#ifndef CONFIG_VENDORLOGGER_SIZE
#define CONFIG_VENDORLOGGER_SIZE    (128*1024)
#endif /* CONFIG_VENDORLOGGER_SIZE */

#define VENDORLOGGER_BUF_SIZE       (512)
#define VENDORLOGGER_HEADER_SIZE    (50)


#define RESETLOG_SETTINGS_INDEX_CLOG    (4)
#define CLOG_VALUE_ENABLED_MASK         (0x00000001)
#define CLOG_VALUE_DISABLED             (0x00000000)
#define CLOG_VALUE_ENABLED              (0x00000001)


struct logger_buffer {
	char data[CONFIG_VENDORLOGGER_SIZE];
	size_t windex;
	size_t wloopcount;
	int wloop;
	struct mutex lock;
	wait_queue_head_t wq;
};

struct logger_info {
	struct logger_buffer* buffer;
	size_t rindex;
	size_t rloopcount;
};

static struct logger_buffer* vendor_logger_buffer = NULL;

ssize_t logger_get_unread_size(struct logger_info* info)
{
	ssize_t ret_size = 0;
	ret_size += (info->buffer->wloopcount - info->rloopcount) * CONFIG_VENDORLOGGER_SIZE;
	ret_size += info->buffer->windex;
	ret_size -= info->rindex;
	return ret_size;
}


ssize_t logger_read(struct logger_buffer* buffer, struct logger_info* info, char* buf, size_t size)
{
	char* data_buf = buf;
	size_t data_size = size;
	size_t read_size = 0;
	size_t unread_size = 0;
	ssize_t ret_size = 0;

	while (0 < data_size) {
		unread_size = logger_get_unread_size(info);
		if (unread_size == 0)
			break;
		if (unread_size > CONFIG_VENDORLOGGER_SIZE) {
			info->rindex = info->buffer->windex;
			info->rloopcount = info->buffer->wloopcount - 1;
			continue;
		}

		read_size = data_size;
		if (read_size > CONFIG_VENDORLOGGER_SIZE - info->rindex)
			read_size = CONFIG_VENDORLOGGER_SIZE - info->rindex;
		if (read_size > unread_size)
			read_size = unread_size;

		memcpy(data_buf, &buffer->data[info->rindex], read_size);
		data_buf += read_size;
		data_size -= read_size;
		ret_size += read_size;
		info->rindex += read_size;
		if (CONFIG_VENDORLOGGER_SIZE <= info->rindex) {
			info->rindex -= CONFIG_VENDORLOGGER_SIZE;
			info->rloopcount++;
		}
	}
	return ret_size;
}


static ssize_t vendorlogger_fop_read(struct file *filp, char __user *buf, size_t count, loff_t *ppos)
{
	struct logger_info* info = (struct logger_info*)filp->private_data;
	char* data = NULL;
	int ret = 0;
	ssize_t size = 0;
	DEFINE_WAIT(wait);

	if ((0 >= count) || (NULL == buf)) {
		pr_err("%s: param error\n", __func__);
		return -EINVAL;
	}
	if (unlikely(!(filp->f_mode & FMODE_READ))) {
		pr_err("%s: f_mode err!!\n", __func__);
		return -EINVAL;
	}
	if (unlikely(NULL == info)) {
		pr_err("%s: logger_info = NULL\n", __func__);
		return -EINVAL;
	}

	while (1) {
		mutex_lock(&info->buffer->lock);
		prepare_to_wait(&info->buffer->wq, &wait, TASK_INTERRUPTIBLE);
		ret = logger_get_unread_size(info);
		mutex_unlock(&info->buffer->lock);
		if (ret != 0)
			break;
		if (filp->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			break;
		}
		if (signal_pending(current)) {
			ret = -EINTR;
			break;
		}
		schedule();
	}
	mutex_lock(&info->buffer->lock);
	finish_wait(&info->buffer->wq, &wait);
	if (0 >= ret) {
		mutex_unlock(&info->buffer->lock);
		pr_err("%s: 0 >= ret(%d)\n", __func__, ret);
		return ret;
	}

	data = vmalloc(count);
	if (unlikely(NULL == data)) {
		mutex_unlock(&info->buffer->lock);
		return -ENOMEM;
	}
	size = logger_read(info->buffer, info, data, count);
	mutex_unlock(&info->buffer->lock);
	if (0 >= size) {
		vfree(data);
		pr_err("%s: size = %zd", __func__, size);
		return -EFAULT;
	}

	if (copy_to_user(buf, data, size)) {
		vfree(data);
		pr_err("%s: error copy_to_user()\n", __func__);
		return -EFAULT;
	}
	vfree(data);
	return size;
}


ssize_t logger_write(struct logger_buffer* buffer, const char* buf, size_t size)
{
	const char* data_buf = buf;
	size_t data_size = size;
	size_t write_size = 0;
	ssize_t ret_size = 0;

	while (0 < data_size) {
		write_size = data_size;
		if (write_size > CONFIG_VENDORLOGGER_SIZE - buffer->windex) {
			write_size = CONFIG_VENDORLOGGER_SIZE - buffer->windex;
		}
		memcpy(&buffer->data[buffer->windex], data_buf, write_size);
		data_buf += write_size;
		data_size -= write_size;
		ret_size += write_size;
		buffer->windex += write_size;
		if (CONFIG_VENDORLOGGER_SIZE <= buffer->windex) {
			buffer->windex = 0;
			buffer->wloop = 1;
			buffer->wloopcount++;
		}
	}
	return ret_size;
}

static ssize_t vendorlogger_write(struct logger_buffer* buffer, const char* buf, size_t size)
{
	struct timespec ts_now;
	struct rtc_time tm;
	char header[VENDORLOGGER_HEADER_SIZE];
	ssize_t pos = 0;
	ssize_t spos = 0;

	ts_now = current_kernel_time();
	ts_now.tv_sec -= sys_tz.tz_minuteswest * 60;
	rtc_time_to_tm(ts_now.tv_sec, &tm);
	snprintf(header, sizeof(header), "%02d-%02d %02d:%02d:%02d.%03lu %5d %5d V "
		,tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, ts_now.tv_nsec / 1000000,
		current->group_leader->pid, current->pid);

	mutex_lock(&buffer->lock);
	for (pos = 0; pos <= size; pos++) {
		if (pos == size || buf[pos] == '\0' || buf[pos] == '\n' || buf[pos] == '\r') {
			if (spos != pos) {
				logger_write(buffer, header, strlen(header));
				logger_write(buffer, &buf[spos], pos - spos);
				logger_write(buffer, "\n", 1);
			}
			spos = pos + 1;
		}
	}
	mutex_unlock(&buffer->lock);

	wake_up_interruptible(&buffer->wq);
	return size;
}

static ssize_t vendorlogger_fop_write(struct file *filp, const char __user *buf, size_t count, loff_t *ppos)
{
	struct logger_info* info = (struct logger_info*)filp->private_data;
	char*  data = NULL;
	ssize_t ret_size = 0;

	if(unlikely(!(filp->f_mode & FMODE_WRITE))) {
		return -EINVAL;
	}

	data = vmalloc(count);
	if (unlikely(NULL == data)) {
		pr_err("%s: NULL == vmalloc(count)\n", __func__);
		return -ENOMEM;
	}
	if (unlikely(copy_from_user(data, buf, count))) {
		vfree(data);
		pr_err("%s: error copy_from_user()\n", __func__);
		return -EFAULT;
	}

	ret_size = vendorlogger_write(info->buffer, data, count);
	vfree(data);

	return ret_size;
}

static unsigned int vendorlogger_fop_poll(struct file *filp, poll_table *wait)
{
	struct logger_info* info = (struct logger_info*)filp->private_data;

	if (unlikely(!(filp->f_mode & FMODE_READ))) {
		pr_err("%s: f_mode err(FMODE_READ)\n", __func__);
		return -EINVAL;
	}

	if(unlikely(NULL == info)) {
		pr_err("%s: logger_info = NULL\n", __func__);
		return -EINVAL;
	}
	poll_wait(filp, &info->buffer->wq, wait);
	
	return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
}

static int vendorlogger_fop_open(struct inode *inode, struct file *filp)
{
	struct logger_info* info = NULL;

	if (unlikely(NULL == vendor_logger_buffer)) {
		pr_err("%s: vendor_logger_buffer = NULL\n", __func__);
		return -ENXIO;
	}

	info = kzalloc(sizeof(struct logger_info), GFP_KERNEL);
	if (unlikely(NULL == info)) {
		return -ENOMEM;
	}
	info->buffer = vendor_logger_buffer;
	filp->private_data = info;

	mutex_lock(&info->buffer->lock);
	if (0 == info->buffer->wloop) {
		info->rindex = 0;
		info->rloopcount = 0;
	} else {
		info->rindex = info->buffer->windex;
		info->rloopcount = info->buffer->wloopcount - 1;
	}
	mutex_unlock(&info->buffer->lock);

	return 0;
}

static int vendorlogger_fop_release(struct inode *inode, struct file *filp)
{
	struct logger_info* info = (struct logger_info*)filp->private_data;

	if (NULL == info) {
		pr_err("%s: vendorlogger = NULL\n", __func__);
		return -ENXIO;
	}
	kfree(info);
	return 0;
}

static struct file_operations vendorlogger_fops = {
	.owner = THIS_MODULE,
	.read = vendorlogger_fop_read,
	.write = vendorlogger_fop_write,
	.poll = vendorlogger_fop_poll,
	.open = vendorlogger_fop_open,
	.release = vendorlogger_fop_release,
};

static struct miscdevice vendorlogger_dev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "vendorlogger",
	.fops = &vendorlogger_fops,
};


int vendorlogger_write_to_log(const char *tag, const char *fmt, ...)
{
	va_list ap;
	char msg[VENDORLOGGER_BUF_SIZE + 2];
	int ofs = 0;
	int ret = 0;

	if (NULL == vendor_logger_buffer) {
		return -1;
	}

	ret = snprintf(&msg[ofs], VENDORLOGGER_BUF_SIZE, "%-8s: ", tag);
	if (unlikely(0 > ret)){
		pr_err("%s: err snprintf() ret=%d\n", __func__,ret);
		return -1;
	}
	ofs = ofs + ret;

	va_start(ap, fmt);
	ret = vsnprintf(&msg[ofs], VENDORLOGGER_BUF_SIZE - ofs, fmt, ap);
	va_end(ap);
	if (unlikely(0 > ret)) {
		pr_err("%s: err vsnprintf() ret=%d\n", __func__,ret);
		return -1;
	}
	ofs = ofs + ret;
	msg[ofs + 1] = '\n';
	msg[ofs + 2] = '\0';
	ofs += 2;

	return vendorlogger_write(vendor_logger_buffer, msg, ofs);
}

static int __init vendorlogger_init(void)
{
	int ret = 0;
	unsigned long value = 0;

	ret = get_emmc_info_settings(RESETLOG_SETTINGS_INDEX_CLOG, &value);
	if (unlikely(ret)) {
		pr_err("%s: failed to get debug settings ret = %d\n", __func__, ret);
		return -EINVAL;
	}
	if (likely((value & CLOG_VALUE_ENABLED_MASK) != CLOG_VALUE_ENABLED)) {
		pr_err("%s: vendorlogger is disable value = %ld\n", __func__, value);
		return -EINVAL;
	}

	vendor_logger_buffer = vzalloc(sizeof(struct logger_buffer));
	if (unlikely(NULL == vendor_logger_buffer)) {
		return -ENOMEM;
	}
	mutex_init(&vendor_logger_buffer->lock);
	init_waitqueue_head(&vendor_logger_buffer->wq);

	ret = misc_register(&vendorlogger_dev);
	if(unlikely(ret)) {
		pr_err("fail to misc_register (MISC_DYNAMIC_MINOR)\n");
		vfree(vendor_logger_buffer);
		return ret;
	}
	return ret;
}

static void __exit vendorlogger_exit(void)
{
	if (NULL != vendor_logger_buffer) {
		vfree(vendor_logger_buffer);
		misc_deregister(&vendorlogger_dev);
	}
}

MODULE_LICENSE("GPL");

late_initcall(vendorlogger_init);
module_exit(vendorlogger_exit);
